from django.shortcuts import render
from django.contrib.auth.decorators import login_required
from django.urls import reverse
from django.http import HttpResponseRedirect, JsonResponse, Http404
from django.contrib.auth import logout, get_user_model
from .models import *
from django.db.models import Count
from django.core.exceptions import PermissionDenied
from .utils import create_room

@login_required
def index(request):
	return render(request, 'django_chatter/index.html')

# This fetches a chatroom given the room ID if a user diretly wants to access the chat.
@login_required
def chatroom(request, uuid):
	user = get_user_model().objects.get(username=request.user)
	room = Room.objects.get(id=uuid)
	if room:
		if user in room.members.all():
			latest_messages = room.message_set.all().order_by('-id')[:50]
			for message in latest_messages:
				message.recipients.add(user)
			return render(request, 'django_chatter/chat-window.html',
				{'room_uuid_json': uuid,
				'latest_messages': latest_messages,
				'room_name': room.__str__()}
				)
		else:
			raise Http404("Sorry! What you're looking for isn't here.")
	else:
		raise Http404("Sorry! What you're looking for isn't here.")

#The following functions deal with AJAX requests
@login_required
def users_list(request):
	users = list(get_user_model().objects.values_list('username', flat = True))
	users_list_json = {'userslist': users}
	return JsonResponse(users_list_json)

@login_required
def get_chat_url(request):
	rooms_with_member_count = Room.objects.annotate(num_members = Count('members'))
	user = get_user_model().objects.get(username=request.user)
	target_user = get_user_model().objects.get(username=request.POST.get('target_user'))

	'''
	AI-------------------------------------------------------------------
		Use the util room creation function to create room for one/two
		user(s). This can be extended in the future to add multiple users
		in a group chat.
	-------------------------------------------------------------------AI
	'''
	if (user == target_user):
		room_id = create_room([user])
	else:
		room_id = create_room([user, target_user])
	new_room_id_json={'room_url': room_id}
	return JsonResponse(new_room_id_json)



from .models import *

def get_chatroom_list(request):
	if request.user:
		if request.user.is_authenticated:
			# Get the list of rooms the user is in, from latest to earliest
			rooms_list = Room.objects.filter(members=request.user).order_by('-date_modified')
			rooms_with_unread = []
			# Go through each list of rooms and check if any of the latest 50 messages 
			# was unread
			for room in rooms_list:
				for message in room.message_set.all().order_by('-id')[:50]:
					if request.user not in message.recipients.all():
						rooms_with_unread.append(room.id)
						break

			#unread rooms are marked with bold letterings
			return ({'rooms_list': rooms_list, 'rooms_with_unread': rooms_with_unread})
		else:
			return ({})
	else:
		return ({})



from channels.generic.websocket import AsyncWebsocketConsumer
import json
from .models import *
from django.contrib.auth import get_user_model
from channels.db import database_sync_to_async
import bleach
from uuid import UUID

#Time libraries used to record the time when the user disconnects.
#New messages will be derived from this time in the view.
from django.utils.timezone import now


'''
AI-------------------------------------------------------------------
    Database Access methods below
-------------------------------------------------------------------AI
'''
@database_sync_to_async
def get_room(room_id, multitenant=False, schema_name=None):
    if multitenant:
        if not schema_name:
            raise AttributeError("Multitenancy support error: \
                scope does not have multitenancy details added. \
                did you forget to add ChatterMTMiddlewareStack to your routing?")
        else:
            from django_tenants.utils import schema_context
            with schema_context(schema_name):
                return Room.objects.get(id=room_id)
    else:
        return Room.objects.get(id=room_id)


'''
AI-------------------------------------------------------------------
    1. Select the Room
    2. Select the user who sent the message
    3. Select the message to be saved
    4. Save message
    5. Set room update time to message date_modified
-------------------------------------------------------------------AI
'''
@database_sync_to_async
def save_message(room, sender, text, multitenant=False, schema_name=None):
    if multitenant:
        if not schema_name:
            raise AttributeError("Multitenancy support error: \
                scope does not have multitenancy details added. \
                did you forget to add ChatterMTMiddlewareStack to your routing?")
        else:
            from django_tenants.utils import schema_context
            with schema_context(schema_name):
                new_message = Message(room=room, sender=sender, text=text)
                new_message.save()
                room.date_modified = new_message.date_modified
                room.save()
    else:
        new_message = Message(room=room, sender=sender, text=text)
        new_message.save()
        room.date_modified = new_message.date_modified
        room.save()


class ChatConsumer(AsyncWebsocketConsumer):

    @database_sync_to_async
    def save_message(self, room, user, message):
        '''
        AI-------------------------------------------------------------------
            1. Select the Room
            2. Select the user who sent the message
            3. Select the message to be saved
            4. Save message
            5. Set room update time to message date_modified
        -------------------------------------------------------------------AI
        '''
        room = room
        sender = user
        text = message
        new_message = Message(room=room, sender=sender, text=text)
        new_message.save()
        room.date_modified = new_message.date_modified
        room.save()

    '''
    AI-------------------------------------------------------------------
        WebSocket methods below
    -------------------------------------------------------------------AI
    '''
    async def connect(self):
        self.user = self.scope['user']

        self.schema_name = self.scope.get('schema_name', None)
        self.multitenant = self.scope.get('multitenant', False)
        for param in self.scope['path'].split('/'):
            try:
                room_id = UUID(param, version=4)
                break
            except ValueError:
                pass
        try:
            self.room = await get_room(room_id, self.multitenant, self.schema_name)
            if self.multitenant:
                from django_tenants.utils import schema_context
                with schema_context(self.schema_name):
                    if self.user in self.room.members.all():
                        room_group_name = 'chat_%s' % self.room.id
                        await self.channel_layer.group_add(
                            room_group_name,
                            self.channel_name
                        )
                        await self.accept()
                    else:
                        await self.disconnect(403)
            else:
                if self.user in self.room.members.all():
                    room_group_name = 'chat_%s' % self.room.id
                    await self.channel_layer.group_add(
                        room_group_name,
                        self.channel_name
                    )
                    await self.accept()
                else:
                    await self.disconnect(403)
        except Exception as ex:
            raise ex
            await self.disconnect(500)

    async def disconnect(self, close_code):
        room_group_name = 'chat_%s' % self.room.id
        await self.channel_layer.group_discard(
            room_group_name,
            self.channel_name
        )

    async def receive(self, text_data):
        username = self.user.username

        text_data_json = json.loads(text_data)
        message = text_data_json['message']
        room_id = text_data_json['room_id']

        # Clean code off message if message contains code
        self.message_safe = bleach.clean(message)
        message_harmful = (self.message_safe != message)

        try:
            # room = await self.get_room(room_id)
            room_group_name = 'chat_%s' % room_id
        except Exception as ex:
            raise ex
            await self.disconnect(500)

        await save_message(self.room,
                                self.user,
                                self.message_safe,
                                self.multitenant,
                                self.schema_name
                                )

        if message_harmful:
            warning = "Your message has been escaped due to security reasons.\
             For more information, see \
             https://en.wikipedia.org/wiki/Cross-site_scripting"
        else:
            warning = ''

        await self.channel_layer.group_send(
            room_group_name,
            {
                'type': 'send_to_websocket',
                'message': self.message_safe,
                'warning': warning,
                'sender': username,
                'room_id': room_id,
            }
        )

    async def send_to_websocket(self, event):
        message = event['message']
        warning = event['warning']
        sender = event['sender']
        room_id = event['room_id']
        if warning == '':
            await self.send(text_data=(json.dumps({
                'message': message,
                'sender': sender,
                'room_id': room_id,
                })))
        else:
            await self.send(text_data=(json.dumps({
                'message': message,
                'sender': sender,
                'warning': warning,
                'room_id': room_id
                })))



import uuid

from django.conf import settings
from django.db import models


class UserProfile(models.Model):
    user = models.OneToOneField(settings.AUTH_USER_MODEL,
            on_delete=models.CASCADE, related_name='profile')
    last_visit = models.DateTimeField()

# This model is used to give date and time when a message was created/modified.
class DateTimeModel(models.Model):
    date_created = models.DateTimeField(auto_now_add=True)
    date_modified = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True

class Room(DateTimeModel):
    id = models.UUIDField(primary_key=True,
            default=uuid.uuid4,
            editable=False)
    members = models.ManyToManyField(settings.AUTH_USER_MODEL)

    def __str__(self):
        memberset = self.members.all()
        members_list = []
        for member in memberset:
            members_list.append(member.username)

        return ", ".join(members_list)

class Message(DateTimeModel):
    sender = models.ForeignKey(settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE, related_name='sender')
    room = models.ForeignKey(Room, on_delete=models.CASCADE)
    text = models.TextField()
    recipients = models.ManyToManyField(settings.AUTH_USER_MODEL,
        related_name='recipients')

    def __str__(self):
        return '"{}" \
        sent by "{}" \
        in Room "{}"'.format(
                        self.text,
                        self.sender,
                        self.room
                    )
